/*******************************************************************************
 * Copyright (c) 2023, Jan Koester jan.koester@gmx.net
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 * Neither the name of the <organization> nor the
 *      names of its contributors may be used to endorse or promote products
 *      derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *******************************************************************************/

#include <iostream>
#include <algorithm>

#include <htmlpp/html.h>

#include <conf.h>

#include <conf.h>
#include <editor.h>
#include <plugin.h>
#include <database.h>

namespace blogi {
    class Template : public PluginApi {
    public:
        Template(){

        };

        ~Template(){

        };

        void initPlugin(){
            SQL sql;
            DBResult res;
            sql << "CREATE TABLE IF NOT EXISTS templates ("
                << "id integer PRIMARY KEY " << Args->database[0]->autoincrement() << ","
                << "layer varchar(255),"
                << "html text,"
                << "type integer"
                << ");";
            Args->database[0]->exec(&sql,res);
        };

        void Rendering(const int tid,libhttppp::HttpRequest *req,libhtmlpp::HtmlElement* curpage,int type){

            blogi::SQL sql;
            blogi::DBResult res;
            sql << "select id,layer,html,type from templates";

            std::string turl=req->getRequest();
            if(turl.rfind('?')>0){
                turl=turl.substr(0,turl.rfind('?'));
            }

            int n = Args->database[tid]->exec(&sql,res);
            if(n<1){
                return;
            }

            for (int i = 0; i < n; i++) {
                if(atoi(res[i][3])==RenderingType::All || type==atoi(res[i][3])){
                    libhtmlpp::HtmlString buf;
                    buf << res[i][2];
                    libhtmlpp::HtmlElement *el=curpage->getElementbyID(res[i][1]);
                    if(el)
                        el->appendChild(buf.parse());
                }
            }

        }

        const char * getAuthor(){
            return "Jan Koester";
        }

        const char * getName(){
            return "template";
        }

        const char * getVersion(){
            return "20241020";
        }

    };


    extern "C" PluginApi* create() {
        return new Template();
    }

    extern "C" void destroy(PluginApi* p) {
        delete p;
    }
};
