/*******************************************************************************
Copyright (c) 2024, Jan Koester jan.koester@gmx.net
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the <organization> nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*******************************************************************************/

#include <vector>
#include <htmlpp/html.h>

#include "session.h"

#pragma once

namespace blogi {
  class CookieBanner {
  public:
    void addRule(const char *pref,const char* cookname,const char* dspname,const char* desc,bool req);

    bool getPermission(const char *sessionid,const char *prefix,const char *cookiename);

    void print(libhtmlpp::HtmlElement &page);
  protected:
    CookieBanner(){};
    ~CookieBanner(){};
    void setPermission(const char *sessionid,const char *prefix,const char *cookiename,bool perm);
    Session session;

    struct Rule{
      Rule(){};
      Rule(const Rule &src);
      Rule &operator=(const Rule &src);
      std::vector<char> name;
      std::vector<char> prefix;
      std::vector<char> displayname;
      std::vector<char> description;
      bool              required;
    };

    std::vector<Rule> Rules;
    libhtmlpp::HtmlElement CookieEl;
  };
};
